import AutoRPE.UtilsRPE.Error as Error
from AutoRPE.UtilsRPE.ReadSourceFile import ReadSourceFile


class Variable:
    def __init__(self, procedure, _name, _type, _dimension, _module=None):
        self.id = None
        self.module = _module
        self.procedure = procedure
        self.name = _name
        self.type = _type
        self.dimension = _dimension
        self.is_dummy_argument = False
        self.intent = None
        self.is_optional = None
        self.is_retval = False
        self.is_pointer = False
        self.position = None
        self.allocatable = False
        self.appears_in_namelist = False
        self.precision = None
        self.is_parameter = None
        self.is_member_of = None
        self.mutable = None
        self.same_as = []

    def set_precision(self, precision):
        assert isinstance(precision, int), "Precision argument in set_precision must be an integer"
        self.precision = precision


class Procedure:
    def __init__(self, _name, module):
        self.name = _name
        self.module = module
        self.variables = []
        self.precision = None
        self.interface = None
        self.dummy_arguments = None
        self.uses = []
        self.accessible_var = {}
        self._variables_dictionary = {}

    def add_variable(self, _variable):
        try:
            assert isinstance(_variable, Variable)
        except AssertionError:
            pass
        _variable.procedure = self
        self.variables.append(_variable)

    def get_variable_by_position(self, position):
        for variable in self.variables:
            if variable.is_dummy_argument:
                if variable.position == position:
                    return variable
        raise Error.VariableNotFound("Variable not found in position %d" % position)

    def get_variable_by_name(self, name):
        # In the pre-processing phase this dictionary is empty
        if self._variables_dictionary:
            try:
                return self._variables_dictionary[name]
            except KeyError:
                None
        else:
            for variable in self.variables:
                if variable.name.lower() == name.lower():
                    return variable

    def fill_dictionary_of_variable(self):
        for var in self.variables:
            self._variables_dictionary[var.name] = var

    def fill_dictionary_of_accessible_var(self):

        if isinstance(self, DerivedType):
            # TODO probably for extended structures I have to add the extension variables
            for var in self.variables:
                self.accessible_var[var.name] = [var]
            return

        for var in self.variables + self.module.variables + self.module.accessible_var:
            try:
                self.accessible_var[var.name] += [var]
            except KeyError:
                self.accessible_var[var.name] = [var]
        try:
            # Main, Interface and TYPE do not have this attribute
            for var in self.use_only_variables:
                try:
                    self.accessible_var[var.name] += [var]
                except KeyError:
                    self.accessible_var[var.name] = [var]
        except AttributeError:
            pass

    def get_dummy_arguments(self):
        if self.dummy_arguments is None:
            number_of_dummy_arguments = len([True for var in self.variables if var.is_dummy_argument])
            self.dummy_arguments = [self.get_variable_by_position(i) for i in range(number_of_dummy_arguments)]
        return self.dummy_arguments


class DerivedType(Procedure):
    def __init__(self, _name, module, _extended_structure=None, ):
        super(DerivedType, self).__init__(_name, module)
        self.extended_structure = _extended_structure
        self.extension = None

    def get_variable(self, item):
        for variable in self.variables:
            if variable.name.lower() == item.lower():
                return variable


class Module(ReadSourceFile):
    def __init__(self, file_name, text, lines):
        super(Module, self).__init__(file_name)
        self.name = self.module_name
        self.main = None
        self.repo = None
        self.procedures = []
        self.variables = []
        self.precision = None
        self.uses = []
        self.used_by = []
        self.accessible_var = []
        self.indirect_accessible_var = []
        self.text = text
        self.lines = lines

    def add_procedure(self, _routine):
        assert isinstance(_routine, Procedure)
        _routine.module = self
        self.procedures.append(_routine)

    def add_variable(self, _variable):
        assert isinstance(_variable, Variable)
        _variable.procedure = self
        self.variables.append(_variable)

    def __getitem__(self, item):
        for _routine in self.procedures:
            if _routine.name == item:
                return _routine
        new_routine = SubRoutine(None, item)
        new_routine.module = self
        self.repo.add_procedure(new_routine)
        self.procedures.append(new_routine)
        return self[item]

    def set_precision(self, precision):
        assert isinstance(precision, int), "Precision argument in set_precision must be an integer"
        for routine in self.procedures:
            routine.set_precision(precision)
        self.precision = precision


class ModuleMain(Procedure):
    def __init__(self, module):
        super(ModuleMain, self).__init__("main", module)
        self.module = module


class Interface:
    def __init__(self, _name, module):
        self.name = _name
        self.module = module
        self.subprogram = []

    def add_procedure(self, subprogram):
        assert isinstance(subprogram, SubRoutine) or isinstance(subprogram, Function)

        self.subprogram.append(subprogram)
        # Don't know it will be useful yet
        if not subprogram.interface:
            subprogram.interface = self
        elif isinstance(subprogram.interface, list):
            subprogram.interface.append(self)
        else:
            subprogram.interface = [subprogram.interface, self]


class SubRoutine(Procedure):
    def __init__(self, _name, module):
        super(SubRoutine, self).__init__(_name, module)
        # External subroutine will just have this attribute to true
        self.is_external = False
        # Case in which a use statement is found inside a subroutine definition
        self.use_only_variables = []


class ExternalSubprogram:
    def __init__(self, name):
        self.name = name
        self.is_external = True


class Function(Procedure):
    def __init__(self, _name, module):
        super(Function, self).__init__(_name, module)
        self.is_external = False
        self.type = None
        self.use_only_variables = []


class Argument:
    def __init__(self, name, type=None, var=None):
        self.name = name
        self.type = type
        self.variable = var


class SubprogramCall:
    def __init__(self, call, subprogram):
        self.call = call
        self.subprogram = subprogram
        self.arguments = []
        self.dummy_arguments = []
