import AutoRPE.UtilsRPE.AttributeParser as AttributeParser
import AutoRPE.UtilsRPE.BasicStructures as BasicStructures
import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
import AutoRPE.UtilsRPE.Splitter as Splitter
import AutoRPE.UtilsRPE.CallManager as CallManager
import AutoRPE.UtilsRPE.NatureDeterminer as NatureDeterminer
import AutoRPE.UtilsRPE.PointerDealer as PointerDealer
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import AutoRPE.UtilsRPE.Error as Error
import AutoRPE.UtilsRPE.RegexPattern as RegexPattern

import re
import warnings


def get_variable(contents, procedure=None, vault=None):
    contents = contents.replace(" ", "").strip()
    contents = BasicFunctions.remove_literals(contents)

    try:
        purged_contents = NatureDeterminer.strip_from_intrinsic(contents, vault)
        if len(purged_contents) > 1:
            raise Error.VariableNotFound("The string %s is not a variable" % contents)
        contents = purged_contents[0]
    except ValueError:
        print("Not able to clean content ", contents, " from intrinsics")
    contents = _clean_contents(contents, vault)

    if Splitter.split_elements_in_comparison(contents):
        raise Error.VariableNotFound("The string %s is not a variable" % contents)

    if NatureDeterminer.has_operations(contents):
        raise Error.VariableNotFound("The string %s is not a variable" % contents)

    if re.search(r"^\w+$", contents):
        try:
            return vault.get_variable_by_name(var_name=contents, procedure=procedure)
        except Error.VariableNotFound:
            raise Error.VariableNotFound
    if contents.count("%val"):
        contents = contents.replace("%val", "")
        return get_variable(contents, procedure, vault)
    if contents.count("%"):
        try:
            return get_structure_member(contents, procedure, vault)
        except Error.MemberNotFound:
            raise Error.VariableNotFound

    if is_keyword_argument(contents):
        # Removing before first =
        contents = contents[contents.find("=") + 1:].strip()
        return get_variable(contents, procedure, vault)

    raise Error.VariableNotFound("Variable %s: not found" % contents)


def get_real_in_line(line, vault, current_block):
    subprogram_call = get_procedure_inline(line, vault, current_block)
    ret_dummy = []
    ret_var = []
    if not subprogram_call:
        if PointerDealer.is_pointer_assignment(line):
            try:
                ptr, target = PointerDealer.get_pointer_target(line, current_block.procedure, vault)
                if [ptr, target] != [None, None]:
                    VariablePrecision.all_real[ptr.type]
                    return [[ptr], [target]]
            except KeyError:
                pass
    else:
        for s_call in subprogram_call:
            if not s_call.subprogram.is_external:
                for dummy_a, arg in zip(s_call.dummy_arguments, s_call.arguments):
                    if dummy_a.intent != "in":
                        try:
                            VariablePrecision.all_real[dummy_a.type]
                            ret_dummy.append(dummy_a)
                            ret_var.append(arg.variable)
                        except KeyError:
                            continue
        return [ret_dummy, ret_var]

    return [None, None]


def get_procedure_inline(line, vault, current_block, only_subroutine=False):
    # To store calls and called procedures names
    procedure_calls = []

    # Search for calls to subprograms
    called_subroutines = CallManager.get_subroutine(line, current_block.procedure, vault=vault)

    if called_subroutines:
        for s in called_subroutines:
            procedure_calls.append(BasicFunctions.close_brackets(line[line.index(s.name):]))
    if not only_subroutine:
        called_functions = CallManager.get_function(line, current_block.procedure, vault=vault)
        if called_functions:
            for f in called_functions:
                procedure_calls.append(BasicFunctions.close_brackets(line[line.index(f.name):]))
    else:
        called_functions = []

    # If are there calls to procedures
    subprogram_call = []
    if called_subroutines + called_functions:
        for call, called_subprogram in zip(procedure_calls, called_subroutines + called_functions):
            called_arguments = CallManager.find_call_arguments(call)

            if not called_arguments:
                continue

            if isinstance(called_subprogram, BasicStructures.Interface):
                called_subprogram = get_interface_from_arguments(called_arguments, interface=called_subprogram,
                                                                 procedure=current_block.procedure, vault=vault)

            sub_prog = BasicStructures.SubprogramCall(call, called_subprogram)
            for index, argument in enumerate(called_arguments):

                key = None
                try:
                    if is_keyword_argument(argument):
                        [key, argument] = argument.split("=", 1)
                        key = key.strip()
                        argument = argument.strip()
                        var_type = get_type_of_contents(argument, procedure=current_block.procedure, vault=vault)
                    else:
                        var_type = get_type_of_contents(argument, procedure=current_block.procedure, vault=vault)
                except Error.VariableNotFound:
                    continue

                if not called_subprogram.is_external:
                    if key is not None:
                        dummy_var = called_subprogram.get_variable_by_name(key)
                    else:
                        dummy_var = called_subprogram.get_variable_by_position(index)
                    sub_prog.dummy_arguments.append(dummy_var)

                try:
                    var = get_variable(argument, procedure=current_block.procedure, vault=vault)
                except Error.VariableNotFound:
                    var = None
                arg = BasicStructures.Argument(argument, var_type, var)
                sub_prog.arguments.append(arg)

            subprogram_call.append(sub_prog)

    return subprogram_call


# Retrieve variable corresponding to a structure member
def get_structure_member(contents, procedure, vault):
    root = contents.split("%")[0]
    try:
        root_variable = procedure.accessible_var[root][0]
    except KeyError:
        # Case of procedure or function with a use module, only var statement
        try:
            root_variable = [v for v in
                             procedure.use_only_variables
                             if v.name.lower() == root][0]
        except IndexError:
            raise Error.VariableNotFound
    try:
        root_derived_type = [dt for dt in vault.derived_types if dt.name.lower() == root_variable.type.lower()][0]
    except IndexError:
        try:
            root_derived_type = [dt for dt in vault.external_types if dt.lower() == root_variable.type.lower()][0]
            raise Error.ExternalTypeUsed("The external type %s has been used" % root_derived_type)
        except IndexError:
            raise Error.DerivedTypeNotFound
    layers = contents.split("%")[1:]
    for l in layers:
        try:
            layer = [v for v in root_derived_type.variables if v.name.lower() == l.lower()][0]
        except IndexError:
            # It means that mother type is an extended type
            try:
                layer = [v for v in root_derived_type.extension.variables if v.name.lower() == l.lower()][0]
            except IndexError:
                raise Error.MemberNotFound
            if layer:
                pass
            else:
                raise Error.VariableNotFound("Variable %s not found", contents)

        try:
            root_derived_type = [dt for dt in vault.derived_types if dt.name.lower() == layer.type.lower()][0]
        except IndexError:
            if l == contents.split("%")[-1]:
                pass
            else:
                raise Error.VariableNotFound("Variable %s not found", contents)

    return layer


def get_type_of_contents(contents, procedure=None, vault=None):
    # TODO: Simplify this function
    # I should do a first filter at this point distinguishing if its a string, a simple variable ("\w+")
    # if its a function/array (is_call_to_function is True), is a keyword argument, or other exceptions

    contents = contents.replace(" ", "").strip()
    contents_lower = contents.lower()

    # Basic check
    if contents[0] == "-" or contents[0] == "+":
        contents = contents[1:]

    if contents_lower[0] in ["'", '"'] or contents_lower[:4].strip() == 'trim':
        return "char"
    if contents.count("'") or contents.count("\""):
        return "char"
    if contents_lower.find("nf90_") == 0:
        return "integer"
    if contents_lower.find("set_nf90_") == 0:
        return "integer"
    is_a_comparison = Splitter.split_elements_in_comparison(contents)
    if is_a_comparison is not None:
        return "logical"
    if contents_lower.count(".false.") or contents_lower.count(".true."):
        return "logical"

    # Check if is an array
    try:
        # suppose that all the elements are of the same type
        contents = Splitter.split_hardcoded_array(contents)[0]
        return get_type_of_contents(contents, procedure, vault)
    except Error.IsNotAnArray:
        pass

    # After checking it is not an array we remove parenthesis
    if BasicFunctions.is_all_inside_parenthesis(contents):
        contents = BasicFunctions.remove_outer_brackets(contents)

    # Case with call to a function with array arguments after call foo(a(1:3), b)
    match = re.search(r"(\w* *\([^=]*\)) *(\(.*\))", contents)
    if match:
        contents = match.group(1)

    # Check if is a call to function, and in this case if is external
    is_an_external_function = False
    if CallManager.is_call_to_function(contents, procedure, vault):
        try:
            return get_function_type(contents, procedure, vault)
        except Error.ProcedureNotFound:
            is_an_external_function = True

    if NatureDeterminer.has_operations(contents):
        split_elements = Splitter.split_elements_in_operations(contents, procedure=procedure, vault=vault)
        types = [get_type_of_contents(element, procedure, vault) for element in
                 split_elements]
        return BasicFunctions.combine_types(types)
    if is_keyword_argument(contents):
        # Removing before first =
        contents = contents[contents.find("=") + 1:].strip()
        try:
            return get_type_of_contents(contents, procedure, vault)
        except Error.VariableNotFound:
            warnings.warn('External variable %s used' % contents)
            return None

    # Remove the index from array and DT:
    #   ght_abl <-- ght_abl(2:jpka)
    #   todelay%z1d <-- todelay(ji)%z1d(:)
    contents = _clean_contents(contents, vault)
    if re.match("^" + RegexPattern.number_exp_DPE_notation + "$", contents):
        return VariablePrecision.wp

    if contents.count(","):
        # If there are commas at this point, it must correspond with a temporal array
        # Array elements should have the same type, so we will assume that all the elements
        # have the same type than the first element
        first_element = AttributeParser.attribute_split(contents)[0]
        return get_type_of_contents(first_element, procedure, vault)

    # Checks if it is a number
    num_type = NatureDeterminer.is_a_number(contents)
    if num_type:
        return num_type

    # Check if is a Derived data type
    if contents.count("%"):
        if contents.split("%")[-1] == "val":
            return VariablePrecision.dp
        try:
            member = get_structure_member(contents, procedure, vault)
        except Error.ExternalTypeUsed:
            return None
        return member.type

    # Search if is a simple variable
    if re.search(r"^\w+$", contents):
        try:
            var = vault.get_variable_by_name(var_name=contents, procedure=procedure)
            return var.type
        except Error.VariableNotFound:
            if contents_lower.count("mpi_"):
                return "integer"
            elif contents_lower == "rpe":
                return "rpe_var"
            elif is_an_external_function:
                warnings.warn("FNot able to find type! :(  |%s|" % contents)
                return None
            else:
                try:
                    vault.get_procedure_by_name(contents, procedure.module.name)
                    # A function is being passed as an arg, as in CALL MPI_OP_CREATE(func,[...] )
                    return None
                except Error.ProcedureNotFound:
                    raise Error.VariableNotFound(contents)

    raise Error.TypeOfContent("Type of content for the argument %s: not found", contents)


def get_function_type(function_call, procedure, vault):
    fun_name = CallManager.find_called_function(function_call)
    fun_name = fun_name.lower()
    if fun_name in vault.list_of_intrinsics:
        if fun_name in ["reshape", "min", "max", "sum", "log", "log10", "maxval", "minval", "abs", "sqrt", "maxloc",
                        "minloc", "huge", "tiny", "transfer", "sign"]:
            return get_type_of_contents(CallManager.find_call_arguments(function_call)[0], procedure, vault)
        elif fun_name in ["int", "nint", "count", "size", "mod", "umask", "len", "len_trim", "count"]:
            return "integer"
        elif fun_name in ["real", "float", "sin", "cos", "acos", "exp", "cosh"]:
            if fun_name == "real":
                return get_type_of_real_cast(function_call)

            return VariablePrecision.wp
        elif fun_name in ["cmplx"]:
            return "complex"
        elif fun_name in ["trim", "ctime", "adjustl"]:
            return "char"
        elif fun_name in ["present"]:
            return "logical"
        elif fun_name in ["null", "c_loc"]:
            return "pointer"
        elif fun_name in ["rpe"]:
            return "rpe_var"
        else:
            raise ValueError("This intrinsic should be considered too: %s" % fun_name)
    else:
        try:
            fun = vault.get_procedure_by_name(fun_name, procedure.module.name)
            if isinstance(fun, BasicStructures.Interface):
                return fun.procedures[0].type
            else:
                return fun.type
        except Error.ProcedureNotFound:
            raise Error.ProcedureNotFound


def get_dimension_of_contents(contents, procedure=None, vault=None):
    # TODO: Improve this function to work properly in all cases.
    contents = BasicFunctions.remove_literals(contents)
    contents = contents
    # Its a Char
    if contents.count("'") or contents.count('"') or contents.lower().count("trim"):
        return 0
    if contents.count(".false.") or contents.count(".true."):
        return 0
    if is_keyword_argument(contents):
        return get_dimension_of_contents(contents.split("=")[1], procedure, vault)
    # Matches arrays in the form of (/a,b,c/) or [a,b,c]
    if NatureDeterminer.is_an_hardcoded_array(contents):
        return 1
    # Adding a workaround for casts
    if CallManager.is_call_to_function(contents, procedure, vault):
        fun = CallManager.find_called_function(contents)
        if fun.lower() in ["real", "rpe", "sqrt", "min", "max"]:
            return get_dimension_of_contents(CallManager.find_call_arguments(contents)[0], procedure, vault)
        elif fun.lower() == "sum" or fun.lower() == "maxval":
            if contents.lower().count("dim="):
                dim = get_dimension_of_contents(CallManager.find_call_arguments(contents)[0], procedure, vault)
                if dim:
                    return dim - 1
                else:
                    return dim
            else:
                contents = CallManager.find_call_arguments(contents)[0]
        else:
            # it was not a cast?
            pass
    if NatureDeterminer.has_operations(contents):
        elements = Splitter.split_elements_in_operations(contents)
        dimensions = [get_dimension_of_contents(e, procedure=procedure, vault=vault) for e in elements]
        dimensions = [d for d in dimensions if d is not None]
        if not dimensions:
            dimensions = [0]
        return max(dimensions)
    if NatureDeterminer.is_a_number(contents):
        return 0
    if contents.count(":"):
        return contents.count(":")
        # Case with call to a function with array arguments after

    # Arrays with index specified
    if re.search(r"\w *\([0-9\w]+\)", contents):
        return 0
    try:
        var = get_variable(contents, procedure=procedure, vault=vault)
        # Checking if we have the var + slice specification, like sdjf%fdta(2,2,1,2)
        # then it must be an scalar
        if re.search(r"%s *\(.*\)" % var.name, contents):
            return 0
        return var.dimension
    except AssertionError:
        return None


def get_variable_id(vault, variable_name, procedure_name=None, module_name=None, ):
    # Small function created to easily find the id of variables that appear in the code
    vs = [v for v in vault.variables if v.name == variable_name]

    if module_name is not None:
        vs = [v for v in vs if v.procedure.module.name == module_name]
    if procedure_name is not None:
        vs = [v for v in vs if v.procedure.name == procedure_name]

    if len(vs) == 1:
        print("Variable found! ID: %i" % vs[0].id)
        return vs[0].id
    elif len(vs) > 1:
        print("More than one variable!")
        return None
    else:
        print("Not a single variable found!")
        return None


def get_interface_from_arguments(arguments, interface, procedure=None, vault=None):
    argument_types = [get_type_of_contents(argument, procedure, vault) for argument in arguments]
    argument_dim = [get_dimension_of_contents(argument, procedure, vault) for argument in arguments]
    keyword_arguments = [_x.split("=")[0].strip() if is_keyword_argument(_x) else False for _x in arguments]
    return get_interface_from_arguments_and_info(argument_types, interface, argument_dim, keyword_arguments, vault)


def get_interface_from_arguments_and_info(argument_types, interface, argument_dimensions=None, keyword_arguments=None,
                                          vault=None):
    # This function is intended to take the argument info (types and dimensions) and return the subroutine
    # inside the interface that matches it.
    number_of_arguments = len(argument_types)
    if keyword_arguments is None:
        keyword_arguments = [False for _x in argument_types]
    for procedure in interface.subprogram:
        if not vault:
            dummy_arguments = procedure.get_dummy_arguments()
        else:
            dummy_arguments = [var for var in procedure.variables if var.is_dummy_argument and not var.is_member_of]

        try:
            mandatory_arguments = [argument for argument in dummy_arguments if not argument.is_optional]
        except AttributeError:
            print("A problem with dummy argument in procedure %s", procedure.name)
        if len(dummy_arguments) < number_of_arguments:
            continue
        if len(mandatory_arguments) > number_of_arguments:
            continue
        match = True
        for index, argument_type in enumerate(argument_types):
            if match is False:
                continue
            keyword = keyword_arguments[index]
            if keyword:
                var = procedure.get_variable_by_name(keyword)
                if var:
                    dummy_type = var.type
                    dummy_dimension = var.dimension
                else:
                    dummy_type = None
                    dummy_dimension = None
                    match = False
            else:
                dummy_type = procedure.get_variable_by_position(index).type
                dummy_dimension = procedure.get_variable_by_position(index).dimension
            if dummy_type != argument_type:
                if (dummy_type in VariablePrecision.real_types and argument_type == "rpe_var") or (
                        dummy_type == "rpe_var" and argument_type in VariablePrecision.real_types):
                    pass
                else:
                    match = False
            if argument_dimensions is not None:
                # At this point we'll just ignore char variables.
                if argument_type == "char":
                    continue
                argument_dimension = argument_dimensions[index]
                if argument_dimension is not None and argument_dimension != dummy_dimension:
                    match = False

        if match:
            return procedure
    raise AssertionError("No matching routine found for interface : %s" % interface.name)


def is_keyword_argument(argument):
    pattern = "^[^()'\"=]*=[^=]"
    if re.search(pattern, argument):
        return True
    else:
        return False


def get_type_of_real_cast(string):
    # Leave only the arguments
    arguments = CallManager.find_call_arguments(string)
    if len(arguments) == 1:
        return VariablePrecision.wp
    elif len(arguments) == 2:
        argument, type = arguments
        return VariablePrecision.lookup_table[type]


def manage_loc_exception(called_arguments, vault, subprogram):
    # When we have a mpp_XXXloc exception instead of linking arguments with dummies we'll link the arguments together
    called_vars = []
    for arg in called_arguments:
        try:
            var = get_variable(arg, procedure=subprogram, vault=vault)
            called_vars.append(var)
        except Error.VariableNotFound:
            continue
    argument_variables = []
    dummies = []
    for index1 in range(len(called_vars)):
        for index2 in range(index1 + 1, len(called_vars)):
            argument_variables.append(called_vars[index1])
            dummies.append(called_vars[index2])
    return argument_variables, dummies


def _clean_contents(contents, vault=None):
    if vault:
        intrinsics = vault.list_of_intrinsics
    else:
        intrinsics = ["reshape", "min", "max", "sum", "log", "maxval", "minval", "abs", "size", "nint"]

    if contents[0] == "-" or contents[0] == "+":
        contents = contents[1:]

    if re.match(r"\(/(.*)/\)", contents):
        contents = re.match(r"\(/(.*)/\)", contents).groups(1)[0]

    if contents[0] == "(":
        if "dummy%s" % contents == BasicFunctions.close_brackets("dummy%s" % contents):
            contents = contents[1:-1]

    if re.match(r"^([^\()]*) *\)+ *$", contents):
        contents = re.match(r"^([^\()]*) *\)+ *$", contents).groups(1)[0]

    if re.match(r".*\) *\(.*\)", contents):
        match = re.match(r"(.*\)) *\(.*\)", contents)
        contents = match.groups(0)[0]

    pattern = r"(\w+) *\("
    m = [_m for _m in re.finditer(pattern, contents)]
    if m:
        for mat in reversed(m):
            prefix = contents[:mat.start(1)]
            part_to_modify = contents[mat.start(1):]
            to_replace = BasicFunctions.close_brackets(part_to_modify)
            replacement = to_replace[:to_replace.find("(")]
            if replacement in intrinsics and CallManager.is_call_to_function(contents):
                continue
            contents = "%s%s" % (prefix, part_to_modify.replace(to_replace, replacement))
    if contents.count("(/"):
        # Is temporal array
        contents = contents.replace("(/", "")
        contents = contents.replace("/)", "")
    return contents
